/* 
 * CuImage, a CUDA image processing class library
 * Copyright (C) 2007 Wouter Caarls, MPIbpc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <windows.h>

#include <cutil.h>
#include <cufft.h>
#include <cublas.h>

#include "CuImage.h"

#include "cmm.cu"
#include "CuImageUtils.cu"
#include "CuImagePixel.cu"
#include "CuImageNeighborhood.cu"
#include "CuImageGeometric.cu"
#include "CuImageFFT.cu"
#include "CuImageHistogram.cu"
#include "CuImageLabeling.cu"

//#include "cmm_example.cu"

CuImage::CuImage(CuImage &img) : cmmPagedMemory(img.GetDimensions())
{
  if (img.m_page->m_hostptr || img.m_page->m_devptr)
  {
    img.Lock();
    Lock();

    cudaMemcpy(GetDevicePtr(), img.GetDevicePtr(), GetDimensions().Size()*sizeof(float), cudaMemcpyDeviceToDevice);

    img.Release();
    Release();
  }
}

CuImage &CuImage::operator=(CuImage &rhs)
{
  if (this != &rhs)
  {
    m_page->Free();
    m_page = new cmmPage();
    m_page->Associate(rhs.GetDimensions(), NULL, true);

    if (rhs.m_page->m_hostptr || rhs.m_page->m_devptr)
    {
      rhs.Lock();
      Lock();

      cudaMemcpy(GetDevicePtr(), rhs.GetDevicePtr(), GetDimensions().Size()*sizeof(float), cudaMemcpyDeviceToDevice);

      rhs.Release();
      Release();
    }
  }

  return *this;
}

CuImage &CuImage::SetDataInt8(const unsigned char *const data)
{
  unsigned char *shortdata;

  CUIMAGE_SAFE_CALL(CuImageMalloc((void **)&shortdata, GetDimensions().Size()*sizeof(unsigned char)));
  CUIMAGE_SAFE_CALL(cudaMemcpy(shortdata, data, GetDimensions().Size()*sizeof(unsigned char), cudaMemcpyHostToDevice));

  dim3 threadBlock(CUIMAGE_THREADS);
  dim3 blockGrid(iDivUp(GetDimensions().Size(), CUIMAGE_THREADS));

  Lock();

  CuImagePixelInt8ToFloat<<<blockGrid, threadBlock>>>(shortdata, GetDevicePtr(), GetDimensions().Size());
  CUIMAGE_SAFE_CALL(cudaGetLastError());

  Release();

  CUIMAGE_SAFE_CALL(CuImageFree(shortdata));

  return *this;
}

unsigned char *CuImage::GetDataInt8(void)
{
  unsigned char *data;
  unsigned char *shortdata;

  dim3 threadBlock(CUIMAGE_THREADS);
  dim3 blockGrid(iDivUp(GetDimensions().Size(), CUIMAGE_THREADS));

  CUIMAGE_SAFE_CALL(CuImageMalloc((void **)&shortdata, GetDimensions().Size()*sizeof(unsigned char)));
  data = (unsigned char *) malloc(GetDimensions().Size()*sizeof(unsigned char));

  Lock();

  CuImagePixelFloatToInt8<<<blockGrid, threadBlock>>>(GetDevicePtr(), shortdata, GetDimensions().Size());
  CUIMAGE_SAFE_CALL(cudaGetLastError());

  Release();

  CUIMAGE_SAFE_CALL(cudaMemcpy(data, shortdata, GetDimensions().Size()*sizeof(unsigned char), cudaMemcpyDeviceToHost));
  CUIMAGE_SAFE_CALL(CuImageFree(shortdata));

  return data;
}

CuImage &CuImage::SetDataInt16(const unsigned short *const data)
{
  unsigned short *shortdata;

  CUIMAGE_SAFE_CALL(CuImageMalloc((void **)&shortdata, GetDimensions().Size()*sizeof(unsigned short)));
  CUIMAGE_SAFE_CALL(cudaMemcpy(shortdata, data, GetDimensions().Size()*sizeof(unsigned short), cudaMemcpyHostToDevice));

  dim3 threadBlock(CUIMAGE_THREADS);
  dim3 blockGrid(iDivUp(GetDimensions().Size(), CUIMAGE_THREADS));

  Lock();

  CuImagePixelInt16ToFloat<<<blockGrid, threadBlock>>>(shortdata, GetDevicePtr(), GetDimensions().Size());
  CUIMAGE_SAFE_CALL(cudaGetLastError());

  Release();

  CUIMAGE_SAFE_CALL(CuImageFree(shortdata));

  return *this;
}

unsigned short *CuImage::GetDataInt16(void)
{
  unsigned short *data;
  unsigned short *shortdata;

  dim3 threadBlock(CUIMAGE_THREADS);
  dim3 blockGrid(iDivUp(GetDimensions().Size(), CUIMAGE_THREADS));

  CUIMAGE_SAFE_CALL(CuImageMalloc((void **)&shortdata, GetDimensions().Size()*sizeof(unsigned short)));
  data = (unsigned short *) malloc(GetDimensions().Size()*sizeof(unsigned short));

  Lock();

  CuImagePixelFloatToInt16<<<blockGrid, threadBlock>>>(GetDevicePtr(), shortdata, GetDimensions().Size());
  CUIMAGE_SAFE_CALL(cudaGetLastError());

  Release();

  CUIMAGE_SAFE_CALL(cudaMemcpy(data, shortdata, GetDimensions().Size()*sizeof(unsigned short), cudaMemcpyDeviceToHost));
  CUIMAGE_SAFE_CALL(CuImageFree(shortdata));

  return data;
}
