/* 
 * CuImage, a CUDA image processing class library
 * Copyright (C) 2007 Wouter Caarls, MPIbpc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "CuImage.hu"

__global__ void CuImageGeometricShift(float *in, int xshift, int yshift, float *out, int width, int height)
{
  const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;
  const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;

  if (x < width && y < height)
    out[IMUL(y, width)+x] = in[periodic2D<false>(x+xshift, y+yshift, width, height)];
}

__global__ void CuImageGeometricSlice(float *in, int xstart, int ystart, int inwidth, float *out, int width, int height)
{
  const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;
  const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;

  if (x < width && y < height)
    out[IMUL(y, width)+x] = in[IMUL(ystart+y, inwidth)+xstart+x];
}

__global__ void CuImageGeometricSliceInterp(float *in, float xstart, float ystart, int inwidth, int inheight, float *out, int width, int height)
{
  const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;
  const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;

  if (x < width && y < height)
  {
    const float2 n = {x+xstart, y+ystart};
    int2 i = {(int)n.x, (int)n.y};
    const float2 f = {n.x-i.x, n.y-i.y};
    float p, rp, dp, rdp;

    p = in[periodic2D<false>(i.x, i.y, inwidth, inheight)];
    rp = in[periodic2D<false>(i.x+1, i.y, inwidth, inheight)];
    dp = in[periodic2D<false>(i.x, i.y+1, inwidth, inheight)];
    rdp = in[periodic2D<false>(i.x+1, i.y+1, inwidth, inheight)];

    out[IMUL(y, width)+x] = 
	        (1 - f.x) * ((1 - f.y) * p  +
													  f.y  * dp) +
				  f.x       * ((1 - f.y) * rp  +
													  f.y  * rdp);
  }
}

__global__ void CuImageGeometricAffine(float *in, float a11, float a12, float a13, float a21, float a22, float a23, float *out, int width, int height)
{
  const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;
  const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;

  if (x < width && y < height)
  {
    const float2 n = {a11*x + a12*y + a13, a21*x + a22*y + a23};
    int2 i = {(int)n.x, (int)n.y};
    const float2 f = {n.x-i.x, n.y-i.y};
#if 0
    if (i.x > 0 && i.y > 0 && i.x < width-1 && i.y < height-1)
#else
	  i.x = max(min(i.x, width-2), 1);
	  i.y = max(min(i.y, height-2), 1);
#endif
    out[IMUL(y, width)+x] =
	        (1 - f.x) * ((1 - f.y) * in[i.x   + IMUL((i.y)  , width)]  +
													  f.y  * in[i.x   + IMUL((i.y+1), width)]) +
				  f.x       * ((1 - f.y) * in[i.x+1 + IMUL((i.y)  , width)]  +
													  f.y  * in[i.x+1 + IMUL((i.y+1), width)]);
#if 0
    else
      out[IMUL(y, width)+x] = 0;
#endif

  }
}

__global__ void CuImageGeometricXX(float *out, int width, int height)
{
  const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;
  const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;

  if (x < width && y < height)
    out[IMUL(y, width)+x] = x;
}

__global__ void CuImageGeometricYY(float *out, int width, int height)
{
  const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;
  const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;

  if (x < width && y < height)
    out[IMUL(y, width)+x] = y;
}

__global__ void CuImageGeometricResize(float *in, int xstart, int ystart, int inwidth, int inheight, float *out, int outwidth, int outheight)
{
  const int x = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;
  const int y = IMUL(blockDim.y, blockIdx.y) + threadIdx.y;

  if (x < outwidth && y < outheight)
  {
    if (x >= xstart && x-xstart < inwidth && y >= ystart && y-ystart < inheight)
      out[IMUL(y, outwidth)+x] = in[IMUL(y-ystart, inwidth)+x-xstart];
    else
      out[IMUL(y, outwidth)+x] = 0;
  }
}

CuImage CuImage::operator()(int xshift, int yshift)
{
  CuImage res(GetDimensions());

  dim3 threadBlock(16, 12);
  dim3 blockGrid(iDivUp(res.GetDimensions().Dim(0), threadBlock.x), iDivUp(GetDimensions().Dim(1), threadBlock.y));

  Lock();
  res.Lock();
  
  CuImageGeometricShift<<<blockGrid, threadBlock>>>(GetDevicePtr(), xshift, yshift, res.GetDevicePtr(), res.GetDimensions().Dim(0), res.GetDimensions().Dim(1));
  CUIMAGE_SAFE_CALL(cudaGetLastError());

  Release();
  res.Release();

  return res;
}

CuImage CuImage::operator()(int xstart, int xend, int ystart, int yend)
{
  CuImage res(cmmDims(xend-xstart, yend-ystart));

  dim3 threadBlock(16, 12);
  dim3 blockGrid(iDivUp(res.GetDimensions().Dim(0), threadBlock.x), iDivUp(res.GetDimensions().Dim(1), threadBlock.y));

  Lock();
  res.Lock();

  CuImageGeometricSlice<<<blockGrid, threadBlock>>>(GetDevicePtr(), xstart, ystart, GetDimensions().Dim(0), res.GetDevicePtr(), res.GetDimensions().Dim(0), res.GetDimensions().Dim(1));
  CUIMAGE_SAFE_CALL(cudaGetLastError());

  Release();
  res.Release();

  return res;
}

CuImage slice(CuImage &img, float xstart, float ystart, int width, int height)
{
  CuImage res(cmmDims(width, height));

  dim3 threadBlock(16, 12);
  dim3 blockGrid(iDivUp(res.GetDimensions().Dim(0), threadBlock.x), iDivUp(res.GetDimensions().Dim(1), threadBlock.y));

  img.Lock();
  res.Lock();

  CuImageGeometricSliceInterp<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), xstart, ystart, img.GetDimensions().Dim(0), img.GetDimensions().Dim(1), res.GetDevicePtr(), res.GetDimensions().Dim(0), res.GetDimensions().Dim(1));
  CUIMAGE_SAFE_CALL(cudaGetLastError());

  img.Release();
  res.Release();

  return res;
}

CuImage affine(CuImage &img, float *a)
{
  return affine(img, a[0], a[1], a[2], a[3], a[4], a[5]);
}

CuImage affine(CuImage &img, float a11, float a12, float a13, float a21, float a22, float a23)
{
  CuImage res(img.GetDimensions());

  dim3 threadBlock(16, 12);
  dim3 blockGrid(iDivUp(res.GetDimensions().Dim(0), threadBlock.x), iDivUp(res.GetDimensions().Dim(1), threadBlock.y));

  img.Lock();
  res.Lock();

  CuImageGeometricAffine<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), a11, a12, a13, a21, a22, a23, res.GetDevicePtr(), res.GetDimensions().Dim(0), res.GetDimensions().Dim(1));
  CUIMAGE_SAFE_CALL(cudaGetLastError());

  img.Release();
  res.Release();

  return res;
}

CuImage xx(CuImage &img)
{
  CuImage res(img.GetDimensions());

  dim3 threadBlock(16, 12);
  dim3 blockGrid(iDivUp(res.GetDimensions().Dim(0), threadBlock.x), iDivUp(res.GetDimensions().Dim(1), threadBlock.y));

  res.Lock();

  CuImageGeometricXX<<<blockGrid, threadBlock>>>(res.GetDevicePtr(), res.GetDimensions().Dim(0), res.GetDimensions().Dim(1));
  CUIMAGE_SAFE_CALL(cudaGetLastError());

  res.Release();

  return res;
}

CuImage yy(CuImage &img)
{
  CuImage res(img.GetDimensions());

  dim3 threadBlock(16, 12);
  dim3 blockGrid(iDivUp(res.GetDimensions().Dim(0), threadBlock.x), iDivUp(res.GetDimensions().Dim(1), threadBlock.y));

  res.Lock();

  CuImageGeometricYY<<<blockGrid, threadBlock>>>(res.GetDevicePtr(), res.GetDimensions().Dim(0), res.GetDimensions().Dim(1));
  CUIMAGE_SAFE_CALL(cudaGetLastError());

  res.Release();

  return res;
}

CuImage resize(CuImage &img, const cmm::cmmDims &dims, int xstart, int ystart)
{
  CuImage res(dims);

  dim3 threadBlock(16, 12);
  dim3 blockGrid(iDivUp(res.GetDimensions().Dim(0), threadBlock.x), iDivUp(res.GetDimensions().Dim(1), threadBlock.y));

  img.Lock();
  res.Lock(); 

  CuImageGeometricResize<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), xstart, ystart, img.GetDimensions().Dim(0), img.GetDimensions().Dim(1), res.GetDevicePtr(), res.GetDimensions().Dim(0), res.GetDimensions().Dim(1));
  CUIMAGE_SAFE_CALL(cudaGetLastError());

  res.Release();
  img.Release();

  return res;
}
