/* 
 * CuImage, a CUDA image processing class library
 * Copyright (C) 2007 Wouter Caarls, MPIbpc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "CuImage.hu"

/* *********************************************************************** *
 * *** Device code ******************************************************* *
 * *********************************************************************** */

__global__ void CuImagePixelInt8ToFloat
(   unsigned char *in, float *out, int size
){
    const int stride = blockDim.x * gridDim.x;
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;

    for (int ii=start; ii < size; ii += stride)
			out[ii] = (float) in[ii];
}

__global__ void CuImagePixelFloatToInt8
(   float *in, unsigned char *out, int size
){
    const int stride = blockDim.x * gridDim.x;
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;

    for (int ii=start; ii < size; ii += stride)
			out[ii] = (unsigned char) in[ii];
}

__global__ void CuImagePixelInt16ToFloat
(   unsigned short *in, float *out, int size
){
    const int stride = blockDim.x * gridDim.x;
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;

    for (int ii=start; ii < size; ii += stride)
			out[ii] = (float) in[ii];
}

__global__ void CuImagePixelFloatToInt16
(   float *in, unsigned short *out, int size
){
    const int stride = blockDim.x * gridDim.x;
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;

    for (int ii=start; ii < size; ii += stride)
			out[ii] = (unsigned short) in[ii];
}

__global__ void CuImagePixelFloatToInt32
(   float *in, unsigned int *out, int size
){
    const int stride = blockDim.x * gridDim.x;
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;

    for (int ii=start; ii < size; ii += stride)
			out[ii] = (unsigned int) in[ii];
}

__global__ void CuImagePixelSplit3
(   float *in, float *out1, float *out2, float *out3, int size
){
    const int stride = blockDim.x * gridDim.x;
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;

    for (int ii=start; ii < size; ii += stride)
    {
			out1[ii] = in[IMUL(ii, 3)+0];
			out2[ii] = in[IMUL(ii, 3)+1];
			out3[ii] = in[IMUL(ii, 3)+2];
    }
}

__global__ void CuImagePixelMerge3
(   float *in1, float *in2, float *in3, float *out, int size
){
    const int stride = blockDim.x * gridDim.x;
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;

    for (int ii=start; ii < size; ii += stride)
    {
			out[IMUL(ii, 3)+0] = in1[ii];
			out[IMUL(ii, 3)+1] = in2[ii];
			out[IMUL(ii, 3)+2] = in3[ii];
    }
}

/* *********************************************************************** *
 * *** Functions ********************************************************* *
 * *********************************************************************** */

#define CuImagePixelP2P_I2I(NAME, OP)\
__global__ void NAME\
(   float *in, float *out, int size\
){\
    const int stride = blockDim.x * gridDim.x;\
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;\
\
    for (int ii=start; ii < size; ii += stride)\
			out[ii] = OP(in[ii]);\
}

#define CuImagePixelP2P_IF2I(NAME, OP)\
__global__ void NAME\
(   float *in, float x, float *out, int size\
){\
    const int stride = blockDim.x * gridDim.x;\
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;\
\
    for (int ii=start; ii < size; ii += stride)\
			out[ii] = OP(in[ii], x);\
}\
__global__ void NAME ## _swap\
(   float *in, float x, float *out, int size\
){\
    const int stride = blockDim.x * gridDim.x;\
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;\
\
    for (int ii=start; ii < size; ii += stride)\
			out[ii] = OP(x, in[ii]);\
}

#define CuImagePixelP2P_II2I(NAME, OP)\
__global__ void NAME\
(   float *in1, float *in2, float *out, int size\
){\
    const int stride = blockDim.x * gridDim.x;\
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;\
\
    for (int ii=start; ii < size; ii += stride)\
			out[ii] = OP(in1[ii], in2[ii]);\
}

#define CuImagePixelReduction_I2F(NAME, OP)\
__global__ void NAME\
(   float *in, float *out, int size\
){\
    const int stride = blockDim.x * gridDim.x;\
		const int start  = IMUL(blockDim.x, blockIdx.x) + threadIdx.x;\
\
    __shared__ float accum[CUIMAGE_REDUCE_THREADS];\
\
		accum[threadIdx.x] = 0;\
    for (int ii=start; ii < size; ii += stride)\
			accum[threadIdx.x] = OP(accum[threadIdx.x], in[ii]);\
\
    __syncthreads();\
		if (!threadIdx.x)\
		{\
			float res = 0;\
			for (int ii = 0; ii < blockDim.x; ii++)\
				res = OP(res, accum[ii]);\
\
			out[blockIdx.x] = res;\
		}\
}

/* *********************************************************************** *
 * *** Operators ********************************************************* *
 * *********************************************************************** */

__device__ float CuImagePixelAdd(float x, float y)
{
  return x+y;
}

CuImagePixelP2P_II2I(CuImagePixelAddII2I, CuImagePixelAdd)
CuImagePixelP2P_IF2I(CuImagePixelAddIF2I, CuImagePixelAdd)
CuImagePixelReduction_I2F(CuImagePixelSumI2F, CuImagePixelAdd)

__device__ float CuImagePixelSubt(float x, float y)
{
  return x-y;
}

CuImagePixelP2P_II2I(CuImagePixelSubtII2I, CuImagePixelSubt)
CuImagePixelP2P_IF2I(CuImagePixelSubtIF2I, CuImagePixelSubt)

__device__ float CuImagePixelMul(float x, float y)
{
  return x*y;
}

CuImagePixelP2P_II2I(CuImagePixelMulII2I, CuImagePixelMul)
CuImagePixelP2P_IF2I(CuImagePixelMulIF2I, CuImagePixelMul)

__device__ float CuImagePixelDiv(float x, float y)
{
  return x/y;
}

CuImagePixelP2P_II2I(CuImagePixelDivII2I, CuImagePixelDiv)
CuImagePixelP2P_IF2I(CuImagePixelDivIF2I, CuImagePixelDiv)

__device__ float CuImagePixelGreater(float x, float y)
{
  return x > y;
}

CuImagePixelP2P_II2I(CuImagePixelGreaterII2I, CuImagePixelGreater)
CuImagePixelP2P_IF2I(CuImagePixelGreaterIF2I, CuImagePixelGreater)

__device__ float CuImagePixelGreaterEqual(float x, float y)
{
  return x >= y;
}

CuImagePixelP2P_II2I(CuImagePixelGreaterEqualII2I, CuImagePixelGreaterEqual)
CuImagePixelP2P_IF2I(CuImagePixelGreaterEqualIF2I, CuImagePixelGreaterEqual)

__device__ float CuImagePixelLess(float x, float y)
{
  return x < y;
}

CuImagePixelP2P_II2I(CuImagePixelLessII2I, CuImagePixelLess)
CuImagePixelP2P_IF2I(CuImagePixelLessIF2I, CuImagePixelLess)

__device__ float CuImagePixelLessEqual(float x, float y)
{
  return x <= y;
}

CuImagePixelP2P_II2I(CuImagePixelLessEqualII2I, CuImagePixelLessEqual)
CuImagePixelP2P_IF2I(CuImagePixelLessEqualIF2I, CuImagePixelLessEqual)

__device__ float CuImagePixelEqual(float x, float y)
{
  return x == y;
}

CuImagePixelP2P_II2I(CuImagePixelEqualII2I, CuImagePixelEqual)
CuImagePixelP2P_IF2I(CuImagePixelEqualIF2I, CuImagePixelEqual)

__device__ float CuImagePixelOr(float x, float y)
{
  return x || y;
}

CuImagePixelP2P_II2I(CuImagePixelOrII2I, CuImagePixelOr)
CuImagePixelP2P_IF2I(CuImagePixelOrIF2I, CuImagePixelOr)

__device__ float CuImagePixelAnd(float x, float y)
{
  return x && y;
}

CuImagePixelP2P_II2I(CuImagePixelAndII2I, CuImagePixelAnd)
CuImagePixelP2P_IF2I(CuImagePixelAndIF2I, CuImagePixelAnd)

__device__ float CuImagePixelXOr(float x, float y)
{
  return (x>0) ^ (y>0);
}

CuImagePixelP2P_II2I(CuImagePixelXOrII2I, CuImagePixelXOr)
CuImagePixelP2P_IF2I(CuImagePixelXOrIF2I, CuImagePixelXOr)

__device__ float CuImagePixelBitAnd(float x, float y)
{
  return ((int)x) & ((int)y);
}

CuImagePixelP2P_II2I(CuImagePixelBitAndII2I, CuImagePixelBitAnd)
CuImagePixelP2P_IF2I(CuImagePixelBitAndIF2I, CuImagePixelBitAnd)

__device__ float CuImagePixelBitOr(float x, float y)
{
  return ((int)x) | ((int)y);
}

CuImagePixelP2P_II2I(CuImagePixelBitOrII2I, CuImagePixelBitOr)
CuImagePixelP2P_IF2I(CuImagePixelBitOrIF2I, CuImagePixelBitOr)

__device__ float CuImagePixelBitXor(float x, float y)
{
  return ((int)x) ^ ((int)y);
}

CuImagePixelP2P_II2I(CuImagePixelBitXorII2I, CuImagePixelBitXor)
CuImagePixelP2P_IF2I(CuImagePixelBitXorIF2I, CuImagePixelBitXor)

__device__ float CuImagePixelMax(float x, float y)
{
  return max(x, y);
}

CuImagePixelP2P_II2I(CuImagePixelMaxII2I, CuImagePixelMax)
CuImagePixelP2P_IF2I(CuImagePixelMaxIF2I, CuImagePixelMax)

__device__ float CuImagePixelMin(float x, float y)
{
  return min(x, y);
}

CuImagePixelP2P_II2I(CuImagePixelMinII2I, CuImagePixelMin)
CuImagePixelP2P_IF2I(CuImagePixelMinIF2I, CuImagePixelMin)

__device__ float CuImagePixelSqrt(float x)
{
  return sqrt(x);
}

CuImagePixelP2P_I2I(CuImagePixelSqrtI2I, CuImagePixelSqrt)

__device__ float CuImagePixelLog(float x)
{
  return log(x);
}

CuImagePixelP2P_I2I(CuImagePixelLogI2I, CuImagePixelLog)

__device__ float CuImagePixelAbs(float x)
{
  return abs(x);
}

CuImagePixelP2P_I2I(CuImagePixelAbsI2I, CuImagePixelAbs)

__device__ float CuImagePixelNot(float x)
{
  return !x;
}

CuImagePixelP2P_I2I(CuImagePixelNotI2I, CuImagePixelNot)

CuImagePixelReduction_I2F(CuImagePixelMinI2F, min)
CuImagePixelReduction_I2F(CuImagePixelMaxI2F, max)

/* *********************************************************************** *
 * *** Host code ********************************************************* *
 * *********************************************************************** */

CuImage merge3(CuImage &in1, CuImage &in2, CuImage &in3)
{
  if (in1.GetDimensions() != in2.GetDimensions() ||
      in2.GetDimensions() != in3.GetDimensions())
    throw runtime_error("Image sizes do not match");

  CuImage res(cmmDims(in1.GetDimensions().Dim(0), in1.GetDimensions().Dim(1), in1.GetDimensions().Dim(2)*3));

  dim3 threadBlock(CUIMAGE_THREADS);
  dim3 blockGrid(iDivUp(in1.GetDimensions().Size(), CUIMAGE_THREADS));

  in1.Lock();
  in2.Lock();
  in3.Lock();
  res.Lock();
  CuImagePixelMerge3<<<blockGrid, threadBlock>>>(in1.GetDevicePtr(), in2.GetDevicePtr(), in3.GetDevicePtr(), res.GetDevicePtr(), in1.GetDimensions().Size());
  CUIMAGE_SAFE_CALL(cudaGetLastError());
  in1.Release();
  in2.Release();
  in3.Release();
  res.Release();

  return res;
}

void split3(CuImage &in, CuImage &out1, CuImage &out2, CuImage &out3)
{
  out1 = CuImage(cmmDims(in.GetDimensions().Dim(0), in.GetDimensions().Dim(1), in.GetDimensions().Dim(2)/3));
  out2 = CuImage(out1.GetDimensions());
  out3 = CuImage(out1.GetDimensions());

  dim3 threadBlock(CUIMAGE_THREADS);
  dim3 blockGrid(iDivUp(out1.GetDimensions().Size(), CUIMAGE_THREADS));

  in.Lock();
  out1.Lock();
  out2.Lock();
  out3.Lock();
  CuImagePixelSplit3<<<blockGrid, threadBlock>>>(in.GetDevicePtr(), out1.GetDevicePtr(), out2.GetDevicePtr(), out3.GetDevicePtr(), out1.GetDimensions().Size());
  CUIMAGE_SAFE_CALL(cudaGetLastError());
  in.Release();
  out1.Release();
  out2.Release();
  out3.Release();

  return;
}

/* *********************************************************************** *
 * *** Functions ********************************************************* *
 * *********************************************************************** */

#define CuImagePixelP2P_I2I_Host(NAME, OP)\
CuImage NAME(CuImage &img)\
{\
  CuImage res(img.GetDimensions());\
\
  dim3 threadBlock(CUIMAGE_THREADS);\
  dim3 blockGrid(iDivUp(res.GetDimensions().Size(), CUIMAGE_THREADS));\
\
  img.Lock();\
  res.Lock();\
  OP<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), res.GetDevicePtr(), res.GetDimensions().Size());\
  CUIMAGE_SAFE_CALL(cudaGetLastError());\
  img.Release();\
  res.Release();\
\
  return res;\
}

#define CuImagePixelP2P_IF2I_Host(NAME, OP)\
CuImage NAME(CuImage &img, const float &scalar)\
{\
  CuImage res(img.GetDimensions());\
\
  dim3 threadBlock(CUIMAGE_THREADS);\
  dim3 blockGrid(iDivUp(res.GetDimensions().Size(), CUIMAGE_THREADS));\
\
  img.Lock();\
  res.Lock();\
  OP<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), scalar, res.GetDevicePtr(), res.GetDimensions().Size());\
  CUIMAGE_SAFE_CALL(cudaGetLastError());\
  img.Release();\
  res.Release();\
\
  return res;\
}\
CuImage NAME(const float &scalar, CuImage &img)\
{\
  CuImage res(img.GetDimensions());\
\
  dim3 threadBlock(CUIMAGE_THREADS);\
  dim3 blockGrid(iDivUp(res.GetDimensions().Size(), CUIMAGE_THREADS));\
\
  img.Lock();\
  res.Lock();\
  OP ## _swap <<<blockGrid, threadBlock>>>(img.GetDevicePtr(), scalar, res.GetDevicePtr(), res.GetDimensions().Size());\
  CUIMAGE_SAFE_CALL(cudaGetLastError());\
  img.Release();\
  res.Release();\
\
  return res;\
}

#define CuImagePixelP2P_II2I_Host(NAME, OP)\
CuImage NAME(CuImage &lhs, CuImage &rhs)\
{\
  if (lhs.GetDimensions() != rhs.GetDimensions())\
    throw runtime_error("Image sizes do not match");\
\
  CuImage res(lhs.GetDimensions());\
\
  dim3 threadBlock(CUIMAGE_THREADS);\
  dim3 blockGrid(iDivUp(res.GetDimensions().Size(), CUIMAGE_THREADS));\
\
  lhs.Lock();\
  rhs.Lock();\
  res.Lock();\
  OP<<<blockGrid, threadBlock>>>(lhs.GetDevicePtr(), rhs.GetDevicePtr(), res.GetDevicePtr(), res.GetDimensions().Size());\
  CUIMAGE_SAFE_CALL(cudaGetLastError());\
  lhs.Release();\
  rhs.Release();\
  res.Release();\
\
  return res;\
}

#define CuImagePixelReduction_I2F_Host(NAME, OP, OP2)\
float NAME(CuImage &img)\
{\
	float *interm, res;\
	float h_interm[CUIMAGE_REDUCE_BLOCKS];\
\
	CUDA_SAFE_CALL( CuImageMalloc((void **)&interm, CUIMAGE_REDUCE_BLOCKS*sizeof(float) ));\
\
	dim3 threadBlock(CUIMAGE_REDUCE_THREADS);\
	dim3 blockGrid(CUIMAGE_REDUCE_BLOCKS);\
\
  img.Lock();\
	OP<<<blockGrid, threadBlock>>>(img.GetDevicePtr(), interm, img.GetDimensions().Size());\
  CUIMAGE_SAFE_CALL(cudaGetLastError());\
  img.Release();\
\
	CUDA_SAFE_CALL( cudaMemcpy(h_interm, interm, CUIMAGE_REDUCE_BLOCKS*sizeof(float), cudaMemcpyDeviceToHost) );\
\
  res = 0;\
	for (int ii=0; ii < CUIMAGE_REDUCE_BLOCKS; ii++)\
    res = OP2(res, h_interm[ii]);\
\
  CUDA_SAFE_CALL(CuImageFree(interm));\
\
  return res;\
}

/* *********************************************************************** *
 * *** Operators ********************************************************* *
 * *********************************************************************** */

CuImagePixelP2P_II2I_Host(operator+, CuImagePixelAddII2I)
CuImagePixelP2P_II2I_Host(operator-, CuImagePixelSubtII2I)
CuImagePixelP2P_II2I_Host(operator*, CuImagePixelMulII2I)
CuImagePixelP2P_II2I_Host(operator/, CuImagePixelDivII2I)

CuImagePixelP2P_II2I_Host(operator>, CuImagePixelGreaterII2I)
CuImagePixelP2P_II2I_Host(operator>=, CuImagePixelGreaterEqualII2I)
CuImagePixelP2P_II2I_Host(operator<, CuImagePixelLessII2I)
CuImagePixelP2P_II2I_Host(operator<=, CuImagePixelLessEqualII2I)

CuImagePixelP2P_II2I_Host(operator==, CuImagePixelEqualII2I)

CuImagePixelP2P_II2I_Host(operator&&, CuImagePixelAndII2I)
CuImagePixelP2P_II2I_Host(operator||, CuImagePixelOrII2I)
CuImagePixelP2P_II2I_Host(operator&, CuImagePixelBitAndII2I)
CuImagePixelP2P_II2I_Host(operator|, CuImagePixelBitOrII2I)
CuImagePixelP2P_II2I_Host(operator^, CuImagePixelBitXorII2I)

CuImagePixelP2P_II2I_Host(fmax, CuImagePixelMaxII2I)
CuImagePixelP2P_II2I_Host(fmin, CuImagePixelMinII2I)

CuImagePixelP2P_IF2I_Host(operator+, CuImagePixelAddIF2I)
CuImagePixelP2P_IF2I_Host(operator-, CuImagePixelSubtIF2I)
CuImagePixelP2P_IF2I_Host(operator*, CuImagePixelMulIF2I)
CuImagePixelP2P_IF2I_Host(operator/, CuImagePixelDivIF2I)

CuImagePixelP2P_IF2I_Host(operator>, CuImagePixelGreaterIF2I)
CuImagePixelP2P_IF2I_Host(operator>=, CuImagePixelGreaterEqualIF2I)
CuImagePixelP2P_IF2I_Host(operator<, CuImagePixelLessIF2I)
CuImagePixelP2P_IF2I_Host(operator<=, CuImagePixelLessEqualIF2I)

CuImagePixelP2P_IF2I_Host(operator==, CuImagePixelEqualIF2I)
CuImagePixelP2P_IF2I_Host(operator&&, CuImagePixelAndIF2I)
CuImagePixelP2P_IF2I_Host(operator||, CuImagePixelOrIF2I)
CuImagePixelP2P_IF2I_Host(operator&, CuImagePixelBitAndIF2I)
CuImagePixelP2P_IF2I_Host(operator|, CuImagePixelBitOrIF2I)
CuImagePixelP2P_IF2I_Host(operator^, CuImagePixelBitXorIF2I)

CuImagePixelP2P_IF2I_Host(fmax, CuImagePixelMaxIF2I)
CuImagePixelP2P_IF2I_Host(fmin, CuImagePixelMinIF2I)

CuImagePixelP2P_I2I_Host(sqrt, CuImagePixelSqrtI2I)
CuImagePixelP2P_I2I_Host(log, CuImagePixelLogI2I)
CuImagePixelP2P_I2I_Host(abs, CuImagePixelAbsI2I)
CuImagePixelP2P_I2I_Host(operator!, CuImagePixelNotI2I)

float CuImagePixelAdd_Host(float x, float y)
{
  return x+y;
}

CuImagePixelReduction_I2F_Host(sum, CuImagePixelSumI2F, CuImagePixelAdd_Host);
CuImagePixelReduction_I2F_Host(fmin, CuImagePixelMinI2F, fminf);
CuImagePixelReduction_I2F_Host(fmax, CuImagePixelMaxI2F, fmaxf);

float mean(CuImage &img)
{
  return sum(img)/img.GetDimensions().Size();
}

float var(CuImage &img)
{
  CuImage temp = img-mean(img);
  return sum(temp*temp)/img.GetDimensions().Size();
}
