/*
>> cmm.h
>>
>> Memory manager for CUDA image processing
>> (c) 2007 Wouter Caarls, MPIbpc
*/

#ifndef __CMM_H_INCLUDED
#define __CMM_H_INCLUDED

#ifdef DLLIMPORTEXPORT
#undef DLLIMPORTEXPORT
#endif

#ifdef CMM_DLL
#define DLLIMPORTEXPORT __declspec(dllexport)
#else
#define DLLIMPORTEXPORT __declspec(dllimport)
#endif

#include <vector>
#include <list>

#pragma warning(disable: 4251)

namespace cmm
{
  class cmmDims;

  // cmmBlock, represents a block of an image
  // Currently limited to 4 dimensions
  class DLLIMPORTEXPORT cmmBlock
  {
  private:
    std::vector<int> m_start;
    std::vector<int> m_end;
    int m_size;

  public:
    explicit cmmBlock(int xs=0, int xe=0, int ys=0, int ye=0,
                      int zs=0, int ze=0, int ws=0, int we=0);
    cmmBlock(const cmmDims &dims);
    bool operator==(const cmmBlock &rhs) const;
    bool operator!=(const cmmBlock &rhs) const;
    int Start(int d) const;
    int End(int d) const;
    int Size() const;
    int Dims() const;
  };

  // cmmDims, represents image dimensionality
  // Currently limited to 4 dimensions
  class DLLIMPORTEXPORT cmmDims
  {
  private:
    std::vector<int> m_dims;
    int m_size, dummy;

  public:
    explicit cmmDims(int x=0, int y=0, int z=0, int w=0);
    cmmDims(const cmmBlock &block);
    bool operator==(const cmmDims &rhs) const;
    bool operator!=(const cmmDims &rhs) const;
    int Dim(int d) const;
    int Size() const;
    int Dims() const;
  };

  class cmmPage;

  // cmmPagedMemory, export class for paged device memory
  // As a principle, memory is not allocated unless necessary,
  // and not deallocated unless necessary. This means that if
  // an image is never used on the host, no host memory is
  // allocated for it.
  // Furthermore, deallocated device memory is cached, so calls
  // to cudaMalloc are avoided as much as possible.
  class DLLIMPORTEXPORT cmmPagedMemory
  {
  protected:
    cmmPage *m_page;

  public:
    // Construct a memory block of size /dims/. If /hostptr/ is given,
    // it refers to preallocated memory. If /free/ is true, the memory
    // will be deallocated when the block is destroyed.
    explicit cmmPagedMemory(const cmmDims &dims, float *hostptr=NULL, bool free=true);

    // Destructor
    ~cmmPagedMemory();

    // Lock a part of the block into device memory.
    float *Lock();
    float *Lock(const cmmBlock &block);

    // Release the locked memory. It will not be unloaded unless
    // flushed (either by the user or by the memory manager).
    void Release();

    // Flush the device memory allocated to the block. Cannot
    // be used when the memory is locked.
    void Flush();

    // Get a pointer to the host memory. Automatically
    // flushes the device memory.
    float *GetHostPtr(void);

    // Get dimensions of the entire image.
    cmmDims GetDimensions(void) const;

    // Get a pointer to device memory.
    float *GetDevicePtr(void) const;

    // Get the part of the image currently locked in the device.
    cmmBlock GetBlock(void) const;
  };

  // State statistics

  class DLLIMPORTEXPORT cmmStateStatistics
  {
  public:
    int pages, device_mem, host_mem_free, host_mem_nofree;

  public:
    cmmStateStatistics() : pages(0), device_mem(0), host_mem_free(0), host_mem_nofree(0) {}
  };

  class DLLIMPORTEXPORT cmmStatistics
  {
  public:
    cmmStateStatistics stored, cached, locked, zombie;
    int pages, device_mem, host_mem_free, host_mem_nofree;
    int locks, relocks, hits, kills, spills;

  public:
    char *Report(void);
  };

  typedef std::list<cmmPage*> cmmPageList;
  typedef cmmPageList::iterator PI;

  // cmmState, implements the page state machine
  class DLLIMPORTEXPORT cmmState
  {
  public:
    cmmPageList *m_list;
    cmmPage *m_page;
    PI m_position;

  public:
    bool operator==(const cmmState &rhs) const 
    {
      return (m_list == rhs.m_list);
    }

    bool operator!=(const cmmState &rhs) const 
    {
      return (m_list != rhs.m_list);
    }

    explicit cmmState(cmmPageList *list);
    explicit cmmState(cmmPage *page, cmmState &init);
    ~cmmState();
    void Change(cmmState &to);
    cmmStateStatistics GetStatistics(void);
  };

  // cmmPage, a block of potentially cached memory
  class DLLIMPORTEXPORT cmmPage
  {
    friend class cmmState;

  private:
    cmmState m_state;
    bool m_free;
    int m_locks;

  public:
    float *m_hostptr;
    cmmDims m_dims;

    float *m_devptr;
    cmmBlock m_block;

  public:
    cmmPage();
    ~cmmPage();

    void Associate(const cmmDims &dims, float *hostptr, bool free);
    void Lock(const cmmBlock &block);
    void Release();
    void Flush();
    void Free();
  };
}

DLLIMPORTEXPORT cmm::cmmStatistics cmmGetStatistics(void);

#endif /* __CMM_H_INCLUDED */
